from Type1 import Type1Font
from PSReadable import toReadable
import StringIO
import FindResource
import VirtualFont

"""
import logging
logging.basicConfig(level=logging.DEBUG)
"""

from knuth import knuth

# Caches for keeping loaded TFMetrics and VirtualFonts
TFMCache = dict()
VFCache = dict()

# =====================================================================

"""
	DviFonts are collections of font dimensions in the context of a .dvi file,
	as defined by a FNTDEF command.  Has a:
	
	fontName (e.g. cmr10, corresponds with the file name of the metrics)
	scaledSize (size, in scaled points, to render at)
	designSize (apparently unused for rendering)
	index (fonts in a DVI file are indexed as they are defined; we keep the index)

	From the fontName, we look up its metrics and store them in
		tfMetrics

	Can represent either a real font or a virtual font.

	Units are Knuth units = Knuth points and SP. 
"""
class DviFont:
	def __init__(self, name, sf, d, index):
		self.fontName=name
		self.scaledSize = sf
		self.designSize = d
		self.index = index

		# We load the metrics only on demand.  Virtual fonts can use several different real fonts
		# and it is likely that we won't acutally use all the fonts defined in a given virtual font.
		# It is inefficient to load metrics for fonts we don't actually end up using in the end.
		self.tfMetrics = None
		self.isUsed=False

		# If the font is virtual, we also keep the virtual character definitions
		self.vChars = None
		self.isVirtual = False # We could just chech vChars == None, but code is more readable with this flag

		# It's handy for a renderer to associate a device font with a DviFont, so we give a field for this
		self.deviceFont = None

	def _str_(self):
		return "DviFont %s %d" % (self.fontName, self.designSize)

	def getCharWidth(self, i):
		return(self.tfMetrics.width[i])

	def getCharHeight(self, i):
		return(self.tfMetrics.height[i])

	def getCharDepth(self, i):
		return(self.tfMetrics.depth[i])

	def load(self):
		if TFMCache.has_key( self.fontName ):
			self.tfMetrics = TFMCache[self.fontName ]
		else:
			self.tfMetrics = TFMetrics( self.fontName );
			TFMCache[ self.fontName ] = self.tfMetrics

		if VFCache.has_key( self.fontName ):
			self.vChars = VFCache[self.fontName ]
		else:
			vfpath = FindResource.getVF( self.fontName );
			# print "vfpath =", vfpath
			if(vfpath != None ):
				self.vChars = VirtualFont.VirtualFont(vfpath)
				VFCache[ self.fontName ] = self.vChars
			else:
				pass
		if self.vChars:
			self.isVirtual = True;
			
		self.isUsed = True

class TFMetrics:

	"""A TFM file keeps character widths as fractions of design size, reported as integer multiples of 2^20.
	The TFM class stores these fractions as ints.  It's more convenient to use these fractions
	as floats, which is what TFMetrics does.  For now TFM reads the raw TFM file and stores ints.  
	TFMetrics uses TFM to make metrics stored as floats.  Maybe these should be merged."""

	def __init__(self, fontfile ):
		self.fontfile = fontfile 
		# print "fontfile = ", fontfile
		self.tfm = Tfm(FindResource.getTFM(fontfile))

		self.width = [0] * 256
		for i in range(self.tfm.bc, self.tfm.ec+1):
			w = self.tfm.charWidth[i]*1.0/(1 << 20)
			self.width[i] = w

		self.height = [ 0 ] * 256
		for i in range(self.tfm.bc, self.tfm.ec+1):
			w = self.tfm.charHeight[i]*1.0/(1 << 20)
			self.height[i] = w

		self.depth = [ 0 ] * 256
		for i in range(self.tfm.bc, self.tfm.ec+1):
			w = self.tfm.charDepth[i]*1.0/(1 << 20)
			self.depth[i] = w

	def getCharDims(self, i):
		return([self.width[i], self.height[i], self.depth[i]])

	def getCharWidth(self, i):
		return(self.width[i])

	def getCharHeight(self, i):
		return(self.height[i])

	def getCharDepth(self, i):
		return(self.depth[i])


# -------------------------------------------------------------------

"""
	The structure of a TFM file is described in the notes TFtoPL by Knuth.  
	All I want to do is read character widths.  These are stored in 4-byte segments,
	each representing a ratio of width to design size.  Of the 32 bits in a word,
	12 bits represent data to the left of the `decimal' point, 20 from the right.
	The range is therefore <b>-2048</b> to <b>2048-2<sup>-20</sup></b>.

	<p>
	At the moment, all we need is designSize, character widths, low index, high index.

	<ul>
	<li><a href="tftopl203-208.ps">Knuth's pages on TFM format</a> (in .ps format)
	</ul>
"""

class Tfm:
	"""
		int lf, lh, nw, nh, nd, ni, nl, nk, ne, np

		The widths of characters are held in this array, as binary fractions offset by 
		2^20.  It has length 256. 

		int[] charWidth
		double designSize
		int bc, ec 
	"""

	def __init__(self, fontname):
		# print "---------------------------- looking for TFM ", fontname
		input = open(fontname, "rb")

#		input.seek(0,2)
#		flen=input.tell()
#		input.seek(0,0)
#		fdata = input.read(flen)
#		input.close()
#		input = StringIO.StringIO(fdata)

		self.lf = knuth.getInt(input, 2)
		self.lh = knuth.getInt(input, 2)
		self.bc = knuth.getInt(input, 2)
		self.ec = knuth.getInt(input, 2)
		self.nw = knuth.getInt(input, 2)
		self.nh = knuth.getInt(input, 2)
		self.nd = knuth.getInt(input, 2)
		self.ni = knuth.getInt(input, 2)
		self.nl = knuth.getInt(input, 2)
		self.nk = knuth.getInt(input, 2)
		self.ne = knuth.getInt(input, 2)
		np = knuth.getInt(input, 2)
		self.readHeader(input)
		""" print "nw, nh, nd = " + str(self.nw) + ", " + str(self.nh) + ", " + str(self.nd)
		print "bc, ec = " + str(self.bc) + ", " + str(self.ec) """
		self.charWidth = [0]*256
		self.charHeight = [0]*256
		self.charDepth = [0]*256
		self.readCharInfo(input)
		self.readWidths(input)
		self.readHeights(input)
		self.readDepths(input)
		input.close()

	# 20 bits to right of `decimal' point
	def toFixWord(self, n):
		return(n*1.0/(1 << 20))

	def readHeader(self, input):
		self.chk = knuth.getWord(input)
		self.designSize = self.toFixWord(knuth.getWord(input))
		for i in range(self.lh-2):
			knuth.getWord(input)

	def readCharInfo(self, input):
		# in tex, only a limited number of character widths are allowed
		# they are stored in an array of charwidths, and each character
		# is assigned an index into this array
		# similarly for heights and deptsh,
		# nut here the arrays contain at most 16 values
		self.wi = []
		self.hi = []
		self.di = []
		for i in range(self.bc):
			self.wi.append(0)
			self.hi.append(0)
			self.di.append(0)
		# print "bc, ec = " + str(self.bc) + ", " + str(self.ec)
		for i in range(self.bc, self.ec+1):
			charinfo = []
			for i in range(4):
				charinfo.append(input.read(1))
			
			self.wi.append(ord(charinfo[0]))
			n = ord(charinfo[1])
			# print "i, h, d = " + str(i) + ", " + str(n >> 4) + ", " + str(n & 15)
			self.hi.append(n >> 4)
			self.di.append(n & 15)
			# skip next two bytes for now
		# print "wi =", self.wi
		# print "length =", len(self.wi)

	def readWidths(self, input):
		w = []
		for i in range(self.nw):
			w.append(knuth.getInt(input, 4))
		self.charWidth = []
		for i in range(256):
			self.charWidth.append(0)
		for i in range (self.bc, self.ec+1):
			""" print "i =", i
			print "w =", self.wi[i] """
			self.charWidth[i] = w[self.wi[i]]

	def readHeights(self, input):
		w = []
		for i in range(self.nh):
			w.append(knuth.getInt(input, 4))
		self.charHeight = []
		for i in range(256):
			self.charHeight.append(0)
		for i in range (self.bc, self.ec+1):
			self.charHeight[i] = w[self.hi[i]]

	def readDepths(self, input):
		w = []
		for i in range(self.nd):
			w.append(knuth.getInt(input, 4))
		self.charDepth = []
		for i in range(256):
			self.charDepth.append(0)
		for i in range (self.bc, self.ec+1):
			# print "i, di = " + str(i) + ", " + str(self.di[i])
			self.charDepth[i] = w[self.di[i]]

